"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JavaObjectStream = exports.JavaObject = void 0;
const ObjectClassMap_1 = require("./ObjectClassMap");
const DataStream_1 = require("./DataStream");
const ObjectInputStream_1 = require("./ObjectInputStream");
const FieldDesc_1 = require("./FieldDesc");
const Utf8Decoder_1 = require("./Utf8Decoder");
const STREAM_MAGIC = 0xACED;
const STREAM_VERSION = 5;
const TYPE_BASE = 0x70;
const TYPES = [
    "Null",
    "Reference",
    "ClassDesc",
    "Object",
    "String",
    "Array",
    "Class",
    "BlockData",
    "EndBlockData",
    "Reset",
    "BlockDataLong",
    "Exception",
    "LongString",
    "ProxyClassDesc",
    "Enum",
];
const HANDLE_BASE = 0x7E0000;
var ClassDescFlags;
(function (ClassDescFlags) {
    ClassDescFlags[ClassDescFlags["WriteMethod"] = 1] = "WriteMethod";
    ClassDescFlags[ClassDescFlags["Serializable"] = 2] = "Serializable";
    ClassDescFlags[ClassDescFlags["Externalizable"] = 4] = "Externalizable";
    ClassDescFlags[ClassDescFlags["BlockData"] = 8] = "BlockData";
    ClassDescFlags[ClassDescFlags["Enum"] = 16] = "Enum";
})(ClassDescFlags || (ClassDescFlags = {}));
class JavaObject {
    constructor(classDesc) {
        this.fields = new Map();
        this.annotations = [];
        this.className = classDesc.className;
        this.serialVersionUid = classDesc.serialVersionUid;
    }
}
exports.JavaObject = JavaObject;
class Handle {
    constructor() {
        this.object = null;
    }
}
class JavaObjectStream {
    constructor(file) {
        this.stream = new DataStream_1.DataStream(file);
        this.handles = [];
        if (this.stream.readUint16() !== STREAM_MAGIC) {
            throw new Error("stream magic mismatch");
        }
        if (this.stream.readUint16() !== STREAM_VERSION) {
            throw new Error("stream version mismatch");
        }
    }
    next() {
        const contentType = this.readType();
        if (contentType === "EndBlockData") {
            throw new Error("unexpected 'EndBlockData' at top level");
        }
        const content = this.readContent(contentType);
        if (content instanceof DataStream_1.DataStream) {
            return content;
        }
        return content.object;
    }
    *read() {
        while (!this.stream.isEndOfStream()) {
            yield this.next();
        }
    }
    readObjectFields(fields) {
        const values = new Map();
        for (const field of fields) {
            const value = this.readFromTypeCode(field.typeCode);
            values.set(field.fieldName, value);
        }
        return values;
    }
    newHandle() {
        const handle = new Handle();
        this.handles.push(handle);
        return handle;
    }
    readType() {
        const type = this.stream.readUint8() - TYPE_BASE;
        if (type < 0 || type >= TYPES.length) {
            throw new Error(`type id '${type}' out of range`);
        }
        return TYPES[type];
    }
    readString(long) {
        let length;
        if (long) {
            length = this.stream.readUint64();
            if (length > Number.MAX_SAFE_INTEGER) {
                throw new Error("long string length out of supported range");
            }
            length = Number(length);
        }
        else {
            length = this.stream.readUint16();
        }
        const data = this.stream.readBytes(length);
        return (0, Utf8Decoder_1.decodeUtf8Buffer)(data);
    }
    readContent(type) {
        switch (type) {
            case "Object":
            case "Class":
            case "Array":
            case "String":
            case "LongString":
            case "Enum":
            case "ClassDesc":
            case "ProxyClassDesc":
            case "Reference":
            case "Null":
            case "Exception":
            case "Reset":
                return this.readObject(type);
            case "BlockData":
            case "BlockDataLong":
                return this.readBlockData(type);
        }
    }
    readObject(type) {
        switch (type) {
            case "Object":
                return this.readNewObject();
            case "Class":
                return this.readNewClass();
            case "Array":
                return this.readNewArray();
            case "String":
            case "LongString":
                return this.readNewString(type);
            case "Enum":
                return this.readNewEnum();
            case "ClassDesc":
            case "ProxyClassDesc":
                return this.readNewClassDesc(type);
            case "Reference":
                return this.readPrevObject();
            case "Null":
                return this.readNullReference();
            case "Exception":
                return this.readException();
            case "Reset":
                return this.readReset();
        }
    }
    readObjectWithType() {
        const type = this.readType();
        switch (type) {
            case "Object":
            case "Class":
            case "Array":
            case "String":
            case "LongString":
            case "Enum":
            case "ClassDesc":
            case "ProxyClassDesc":
            case "Reference":
            case "Null":
            case "Exception":
            case "Reset":
                break;
            default:
                throw new Error(`unexpected type '${type}' for object`);
        }
        return this.readObject(type);
    }
    readNewClass() {
        throw new Error("unimplemented");
    }
    readClassDesc(type) {
        switch (type) {
            case "ClassDesc":
            case "ProxyClassDesc":
                return this.readNewClassDesc(type);
            case "Reference":
                return this.readPrevObject();
            case "Null":
                return this.readNullReference();
        }
    }
    readClassDescWithType() {
        const classDescType = this.readType();
        switch (classDescType) {
            case "ClassDesc":
            case "ProxyClassDesc":
            case "Reference":
            case "Null":
                break;
            default:
                throw new Error(`unexpected type '${classDescType}' for classDesc`);
        }
        return this.readClassDesc(classDescType);
    }
    readNewClassDesc(type) {
        if (type === "ProxyClassDesc") {
            throw new Error("unimplemented");
        }
        const className = this.readString();
        const serialVersionUid = this.stream.readUint64();
        const newHandle = this.newHandle();
        const classDescInfo = this.readClassDescInfo();
        newHandle.object = Object.assign({ className,
            serialVersionUid }, classDescInfo);
        return newHandle;
    }
    readClassDescInfo() {
        const flags = this.stream.readUint8();
        const fields = [];
        const fieldCount = this.stream.readUint16();
        for (let i = 0; i < fieldCount; ++i) {
            fields.push(this.readFieldDesc());
        }
        const annotations = this.readAnnotations();
        const superClassDesc = this.readClassDescWithType().object;
        return {
            flags,
            fields,
            annotations,
            superClassDesc,
        };
    }
    readAnnotations() {
        const annotations = [];
        for (;;) {
            const contentType = this.readType();
            if (contentType === "EndBlockData") {
                break;
            }
            const content = this.readContent(contentType);
            if (content instanceof DataStream_1.DataStream) {
                annotations.push(content);
            }
            else {
                annotations.push(content.object);
            }
        }
        return annotations;
    }
    readFieldDesc() {
        const typeCode = String.fromCharCode(this.stream.readUint8());
        const fieldName = this.readString();
        switch (typeCode) {
            case "B":
            case "C":
            case "D":
            case "F":
            case "I":
            case "J":
            case "S":
            case "Z":
                return new FieldDesc_1.PrimitiveFieldDesc(typeCode, fieldName);
            case "[":
            case "L":
                break;
            default:
                throw new Error(`unexpected field typeCode '${typeCode}'`);
        }
        const className = this.readObjectWithType().object;
        if (typeof className !== "string") {
            throw new Error("expected string for field class name");
        }
        return new FieldDesc_1.ObjectFieldDesc(typeCode, fieldName, className);
    }
    readNewArray() {
        const classDesc = this.readClassDescWithType().object;
        if (!classDesc) {
            throw new Error("got null class desc for array");
        }
        const size = this.stream.readUint32();
        const newHandle = this.newHandle();
        const typeCode = classDesc.className.slice(1, 2);
        switch (typeCode) {
            case "B":
            case "C":
            case "D":
            case "F":
            case "I":
            case "J":
            case "S":
            case "Z":
            case "[":
            case "L":
                break;
            default:
                throw new Error(`unexpected array typeCode '${typeCode}'`);
        }
        const values = [];
        for (let i = 0; i < size; ++i) {
            values.push(this.readFromTypeCode(typeCode));
        }
        newHandle.object = values;
        return newHandle;
    }
    readNewObject() {
        const classDesc = this.readClassDescWithType().object;
        if (!classDesc) {
            throw new Error("got null class desc for object");
        }
        const newHandle = this.newHandle();
        const objectClass = (0, ObjectClassMap_1.getObjectClass)(classDesc.className, classDesc.serialVersionUid);
        const newObject = objectClass ? new objectClass() : new JavaObject(classDesc);
        // The classes need re-ordering from superclass to subclass.
        const classDescChain = [];
        for (let i = classDesc; i !== null; i = i.superClassDesc) {
            classDescChain.unshift(i);
        }
        // Walk the class chain to read the class data.
        for (const dataClassDesc of classDescChain) {
            if ((dataClassDesc.flags & ClassDescFlags.Enum) !== 0) {
                throw new Error("unimplemented");
            }
            if ((dataClassDesc.flags & ClassDescFlags.Externalizable) !== 0) {
                throw new Error("unimplemented");
            }
            if ((dataClassDesc.flags & ClassDescFlags.Serializable) === 0) {
                throw new Error("expected the serializable flag to be set");
            }
            if ((dataClassDesc.flags & ClassDescFlags.WriteMethod) === 0) {
                const values = this.readObjectFields(dataClassDesc.fields);
                for (const [fieldName, value] of values) {
                    if (newObject instanceof JavaObject) {
                        newObject.fields.set(fieldName, value);
                    }
                    else {
                        if (fieldName in newObject) {
                            newObject[fieldName] = value;
                        }
                    }
                }
                continue;
            }
            const dataClass = (0, ObjectClassMap_1.getObjectClass)(dataClassDesc.className, dataClassDesc.serialVersionUid);
            if (!dataClass) {
                if (!dataClass) {
                    console.warn(`no class registered for ${dataClassDesc.className} with UID ${dataClassDesc.serialVersionUid}`);
                }
                const values = this.readObjectFields(dataClassDesc.fields);
                const annotations = this.readAnnotations();
                if (newObject instanceof JavaObject) {
                    for (const [fieldName, value] of values) {
                        newObject.fields.set(fieldName, value);
                    }
                    for (const annotation of annotations) {
                        newObject.annotations.push(annotation);
                    }
                }
                continue;
            }
            if (typeof dataClass.prototype.readObject !== "function") {
                throw new Error("registered class does not implement readObject method");
            }
            if (newObject instanceof JavaObject) {
                const tempObject = Object.create(null);
                dataClass.prototype.readObject.call(tempObject, new ObjectInputStream_1.ObjectInputStream(this, tempObject, dataClassDesc.fields));
                for (const property in tempObject) {
                    if (Object.prototype.hasOwnProperty.call(tempObject, property)) {
                        newObject.fields.set(property, tempObject[property]);
                    }
                }
                const annotations = this.readAnnotations();
                for (const annotation of annotations) {
                    newObject.annotations.push(annotation);
                }
                continue;
            }
            // I'm not sure this is strictly sane in JavaScript land.
            dataClass.prototype.readObject.call(newObject, new ObjectInputStream_1.ObjectInputStream(this, newObject, dataClassDesc.fields));
            // Read any leftover annotations, if the readObject implementation has read
            // everything this will just hit the EndBlockData and complete.
            this.readAnnotations();
        }
        newHandle.object = newObject;
        if (!(newObject instanceof JavaObject) && typeof newObject.readResolve === "function") {
            newHandle.object = newObject.readResolve();
        }
        return newHandle;
    }
    readFromTypeCode(typeCode) {
        switch (typeCode) {
            case "B":
                return this.stream.readUint8();
            case "C":
                return String.fromCharCode(this.stream.readUint16());
            case "D":
                return this.stream.readFloat64();
            case "F":
                return this.stream.readFloat32();
            case "I":
                return this.stream.readUint32();
            case "J":
                return this.stream.readUint64();
            case "S":
                return this.stream.readUint16();
            case "Z":
                return this.stream.readUint8() !== 0;
            case "[":
            // TODO: Not sure about this.
            // eslint-disable-next-line no-fallthrough
            case "L":
                // TODO: Should probably be checking the type against field.className?
                return this.readObjectWithType().object;
        }
    }
    readBlockData(type) {
        let length = null;
        switch (type) {
            case "BlockData":
                length = this.stream.readUint8();
                break;
            case "BlockDataLong":
                length = this.stream.readUint32();
                break;
        }
        const bytes = this.stream.readBytes(length);
        return new DataStream_1.DataStream(bytes);
    }
    readNewString(type) {
        const newHandle = this.newHandle();
        newHandle.object = this.readString(type === "LongString");
        return newHandle;
    }
    readNewEnum() {
        const classDesc = this.readClassDescWithType().object;
        if (!classDesc) {
            throw new Error("got null class desc for enum");
        }
        const newHandle = this.newHandle();
        const enumConstantName = this.readObjectWithType().object;
        if (typeof enumConstantName !== "string") {
            throw new Error("expected string for enum constant name");
        }
        newHandle.object = {
            classDesc,
            enumConstantName,
        };
        return newHandle;
    }
    readPrevObject() {
        const handle = this.stream.readUint32() - HANDLE_BASE;
        if (handle < 0 || handle >= this.handles.length) {
            throw new Error(`prevObject handle '${handle}' out of range [0, ${this.handles.length})`);
        }
        return this.handles[handle];
    }
    readNullReference() {
        return new Handle();
    }
    readException() {
        throw new Error("unimplemented");
    }
    // TODO: The way this is called isn't quite right, while we're doing the right thing,
    //       the code should keep reading if it sees a TC_RESET and continue on to the next thing,
    //       as they can appear anywhere in the object stream and should just be ignored
    //       other than resetting the handle list. As we've never seen anything using them yet,
    //       just treat it as unimplemented for now so we can get the right return types elsewhere.
    readReset() {
        throw new Error("unimplemented");
        // This is the right implementation, but we need to fix the caller behavior.
        // this.handles = [];
    }
}
exports.JavaObjectStream = JavaObjectStream;
JavaObjectStream.RegisterObjectClass = ObjectClassMap_1.registerObjectClass;
